/**
 * 
 */
package explorer;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Iterator;

import alice.logictuple.LogicTuple;
import alice.logictuple.Value;
import alice.logictuple.Var;
import alice.tucson.api.AgentId;
import alice.tucson.api.TucsonException;
import alice.tucson.api.TupleCentreId;
import alice.tuprolog.Int;
import alice.tuprolog.InvalidTheoryException;
import alice.tuprolog.Prolog;
import alice.tuprolog.Theory;


/**
 * This class implements an Advanced Explorer Agent
 * @author andreamonaldini
 *
 */
public class AdvancedExplorerAgent extends ExplorerAgent {
	Prolog prolog;	//the prolog engine: the core of the AdvancedAgentExplorer
	FileInputStream advancedExplorerAgentTheory;	//Prolog Theory for the Advanced Explorer Agent
	TupleCentreId tc; //tuple centre for agents coordination
	boolean plan;	//true if the explorer is following a planned path.
	/**
	 * Constructs a new Advanced Explorer Agent 
	 * @param id Agent ID.
	 * @param grid The virual space where agent works.
	 * @param x The x position of the agent
	 * @param y The y position of the agent
	 * @param advancedExplorerAgentTheory Prolog theory for path planning
	 * @throws TucsonException
	 */
	public AdvancedExplorerAgent(AgentId id, SpaceGrid grid, int x, int y, String advancedExplorerAgentTheory) throws TucsonException {
		super(id, grid, x, y);
		try {
			this.advancedExplorerAgentTheory=new FileInputStream(advancedExplorerAgentTheory);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		this.tc=null;
	}
	/**
	 * Constructs a new Advanced Explorer Agent 
	 * @param id Agent ID.
	 * @param grid The virual space where agent works.
	 * @param x The x position of the agent
	 * @param y The y position of the agent
	 * @param advancedExplorerAgentTheory Prolog theory for path planning
	 * @param tc The tuple centre id
	 * @throws TucsonException
	 */
	public AdvancedExplorerAgent(AgentId id, SpaceGrid grid, int x, int y, String advancedExplorerAgentTheory,TupleCentreId tc) throws TucsonException {
		super(id, grid, x, y);
		try {
			this.advancedExplorerAgentTheory=new FileInputStream(advancedExplorerAgentTheory);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		this.tc=tc;
	}
	
	/* (non-Javadoc)
	 * @see alice.tucson.api.Agent#body()
	 */
	@Override
	protected void body() {
		this.prolog=new Prolog();		//new prolog engine used to find a path from current position to destination
		this.plan=false;
		try {
			Theory theory = new Theory(this.advancedExplorerAgentTheory);
			//set the prolog theory and add the agentId to the theory
			prolog.setTheory(theory);
			//prolog.addTheory(new Theory("agentId("+this.id+")"));
			prolog.solve(new alice.tuprolog.Struct("assert",new alice.tuprolog.Struct("agentId",new alice.tuprolog.Struct(this.id.toString()))));
		} catch (InvalidTheoryException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		alice.tuprolog.SolveInfo answer;
		int forbiddenCellX=-1;	//destination cell-X
		int forbiddenCellY=-1;	//destination cell-X
		int newCellX=-1;
		int newCellY=-1;
		int newPosX=-1;
		int newPosY=-1;
		alice.tuprolog.Struct nextStep;	//next Step in the plan
		Iterator iterator=null;
		alice.tuprolog.Struct path;		//path planned
		boolean moved=false;		//true if the explorer 
		boolean firstStep=false;
		boolean haveDestination=false;	//used if the agent has a destination but there aren't yet a path between his position and the destination cell.
		try{
			//tuple centre for agents coordination
			if (tc==null){
				this.tc=new TupleCentreId("tc");
			}
			
			//Initial Syncronization
			rd(this.tc, new LogicTuple("startPositioning"));
			
			//output of the agent position
			//template explorerAt(X, Y, AgentID)
			out(this.tc,new LogicTuple("advancedAgent", new Value(this.id.toString())));
			out(this.tc,new LogicTuple("explorerAt", new Value(this.x),new Value(this.y), new Value(this.id.toString())));
			System.out.println("AdvancedExplorerAgent "+this.id+" placed at "+this.x+","+this.y);
			//the cell where the agent is initially places is already explored...
			//template cell(X,Y)
			rd(this.tc, new LogicTuple("start"));
			out(tc, new LogicTuple("cell", new Value(this.x),new Value(this.y)));
			
			
			
			//######### agent's control flow - the agent's work terminate when all cells are explored ##############
			while(rdp(this.tc, new LogicTuple("virtualSpaceExplored"))==null){
				//Wait for the right round
				in(this.tc,new LogicTuple("resource_token",new Value(id.toString())));
				
				if(!plan){
					if (!haveDestination){
						//if there isn't a plan to be followed the agent wait for a forbiddenCell(X,Y) tuple (until a forbidden cell is found by Simple Explorers)
						//LogicTuple forbiddenCellTuple = in(this.tc, new LogicTuple("forbiddenCell",new Var("X"), new Var("Y")));
						LogicTuple forbiddenCellTuple = inp(this.tc, new LogicTuple("forbiddenCell",new Var("X"), new Var("Y")));
						if(forbiddenCellTuple!=null){
							
							forbiddenCellX=forbiddenCellTuple.getVarValue("X").intValue();
							forbiddenCellY=forbiddenCellTuple.getVarValue("Y").intValue();
							
							//the explorer tells that he has assigned the forbidden cell himself
							out(this.tc, new LogicTuple("forbiddenCellAssigned",new Value(forbiddenCellX), new Value(forbiddenCellY)));
							System.out.println("Forbidden Cell "+forbiddenCellX+","+forbiddenCellY+" has been check-out by "+this.id);
							
							haveDestination=true;
							
							//update the prolog theory with new explored cells (based on ReSpecT reaction: in(cell(X,Y))-->out_r(newFreeCell(X,Y))
							while(rdp(this.tc, new LogicTuple("newCell",new Var("X"), new Var("Y"),new Value(this.id.toString())))!=null){
								//read the newCell(X,Y) tuple
								//System.out.println("sono qui1"+this.id.toString());
								LogicTuple newCellTuple = in(this.tc, new LogicTuple("newCell",new Var("X"), new Var("Y"),new Value(this.id.toString())));
								newCellX=newCellTuple.getVarValue("X").intValue();
								newCellY=newCellTuple.getVarValue("Y").intValue();
								
								//add the tuple cell(X,Y) to the prolog theory
								prolog.solve(new alice.tuprolog.Struct("assert",new alice.tuprolog.Struct("cell",new Int(newCellX),new Int(newCellY))));
							}
//							add the tuple forbiddenCellAssigned(X,Y) to the prolog theory
							prolog.solve(new alice.tuprolog.Struct("assert",new alice.tuprolog.Struct("cell",new Int(forbiddenCellX),new Int(forbiddenCellY))));
							
							//Plan to reach the forbidden cell from the actual position
							answer=prolog.solve("findPath(cell("+this.x+","+this.y+"),cell("+forbiddenCellX+","+forbiddenCellY+"),Path).");
						//	System.out.println("sono qui2"+this.id.toString());
							if (answer.isSuccess()){
								path=(alice.tuprolog.Struct)answer.getTerm("Path");
								iterator = path.listIterator();
//								we have a new plan!!!
								System.out.println("AdvancedExplorerAgent "+this.id+" at "+this.x+","+this.y+" has a plan: "+path.toString());
								//we have a new plan!!!
								this.plan=true;
								firstStep=true;
								moved=false;
							}
						}else{
							//do nothing (there aren't forbiddenCells)
						}
					}else{	//if there is already a destination
						//update the prolog theory with new explored cells (based on ReSpecT reaction: in(cell(X,Y))-->out_r(newFreeCell(X,Y))
						while(rdp(this.tc, new LogicTuple("newCell",new Var("X"), new Var("Y"),new Value(this.id.toString())))!=null){
							//read the newCell(X,Y) tuple
							//System.out.println("sono qui3"+this.id.toString());
							LogicTuple newCellTuple = in(this.tc, new LogicTuple("newCell",new Var("X"), new Var("Y"),new Value(this.id.toString())));
							newCellX=newCellTuple.getVarValue("X").intValue();
							newCellY=newCellTuple.getVarValue("Y").intValue();
							
							//add the tuple cell(X,Y) to the prolog theory
							prolog.solve(new alice.tuprolog.Struct("assert",new alice.tuprolog.Struct("cell",new Int(newCellX),new Int(newCellY))));
						}
						
						//Plan to reach the forbidden cell from the actual position
						answer=prolog.solve("findPath(cell("+this.x+","+this.y+"),cell("+forbiddenCellX+","+forbiddenCellY+"),Path).");
						//System.out.println("sono qui4"+this.id.toString());
						if (answer.isSuccess()){
							path=(alice.tuprolog.Struct)answer.getTerm("Path");
							iterator = path.listIterator();
//							we have a new plan!!!
							System.out.println("AdvancedExplorerAgent "+this.id+" at "+this.x+","+this.y+" has a plan: "+path.toString());
							//we have a new plan!!!
							this.plan=true;
							firstStep=true;
							moved=false;
						}else{
							System.out.println("AdvancedExplorerAgent "+this.id+" at "+this.x+","+this.y+" has a destination ("+forbiddenCellX+","+forbiddenCellY+") but not a plan");
						}
						
					}
				}else{	//there is already a plan
					if (moved||firstStep){
						//if in the round before the agent has moved
						nextStep=(alice.tuprolog.Struct)iterator.next();
						if (firstStep){nextStep=(alice.tuprolog.Struct)iterator.next();}	//the first cell on the path is the curren cell
						newPosX=nextStepX(nextStep);
						newPosY=nextStepY(nextStep);
						//System.out.println("nextStepX:  "+newPosX+"   nextStepY:   "+newPosY);
						firstStep=false;
					}
					
					//if (rdp(tc, new LogicTuple("explorerAt", new Value(newPosX),new Value(newPosY), new Var("_")))==null){
					//the explorer can move
					go(newPosX,newPosY);
					moved=true;
					//}else{
					//	moved=false;
					//	System.out.println(""+this.id+" is waiting room on "+newPosX+","+newPosY);
					//}
					
					if((!iterator.hasNext())&&moved){
						//the plan is ended and the forbiddenCell is free.
						this.plan=false;
						haveDestination=false;
						in(this.tc, new LogicTuple("forbiddenCellAssigned",new Value(forbiddenCellX), new Value(forbiddenCellY)));
						out(this.tc, new LogicTuple("cell",new Value(forbiddenCellX), new Value(forbiddenCellY)));
						prolog.solve(new alice.tuprolog.Struct("retract",new alice.tuprolog.Struct("cell",new Int(forbiddenCellX),new Int(forbiddenCellY))));
						System.out.println("Forbidden Cell "+forbiddenCellX+","+forbiddenCellY+" has been explored by "+this.id);
					}
				}
				out(this.tc,new LogicTuple("resource_token",new Value(id.toString())));
				System.out.flush();
			}
			//remove the last position tuple 
			in(this.tc, new LogicTuple("explorerAt", new Value(this.x),new Value(this.y), new Value(this.id.toString())));
			System.out.println("AdvancedExplorerAgent "+this.id+" exited - total explored cells = "+
					rdp(this.tc, new LogicTuple("exploredCell",new Var("X"))).getVarValue("X").intValue());
		}catch(Exception e){
			System.out.println(prolog.getTheory().toString());
			System.out.println("AdvancedExplorerAgent - Control error");
		}
		
		
	}
	
	private int nextStepX(alice.tuprolog.Struct s){
		int newX=-1;
		try {
			alice.tuprolog.Var varX=new alice.tuprolog.Var("X");
			alice.tuprolog.Var varY=new alice.tuprolog.Var("Y");
			alice.tuprolog.Struct str=new alice.tuprolog.Struct("cell",varX,varY);
			s.unify(str);	//using unification as a powerful string tokenyzer
			alice.tuprolog.Term termX=varX.copy();
			newX = ((alice.tuprolog.Number)(termX)).intValue();
		}catch(Exception e){System.out.println("AdvancedExplorerAgent - Error adding or removing position tuple.");}
		return newX;
		
	}	
	private int nextStepY(alice.tuprolog.Struct s){
		int newY=-1;
		try {
			alice.tuprolog.Var varX=new alice.tuprolog.Var("X");
			alice.tuprolog.Var varY=new alice.tuprolog.Var("Y");
			alice.tuprolog.Struct str=new alice.tuprolog.Struct("cell",varX,varY);
			s.unify(str);	//using unification as a powerful string tokenyzer
			alice.tuprolog.Term termY=varY.copy();
			newY = ((alice.tuprolog.Number)(termY)).intValue();
		}catch(Exception e){System.out.println("AdvancedExplorerAgent - Error adding or removing position tuple.");}
		return newY;
		
	}
	
	private void go(int newX, int newY){
		try {
			
			//remove the old position tuple explorerAt(X,Y,AgentId)
			in(this.tc, new LogicTuple("explorerAt", new Value(this.x),new Value(this.y), new Value(this.id.toString())));	
			this.x=newX;
			this.y=newY;
//			add the new position tuple explorerAt(X,Y,AgentId)
			out(this.tc,new LogicTuple("explorerAt", new Value(this.x),new Value(this.y), new Value(this.id.toString())));
			System.out.println("AdvancedExplorerAgent "+this.id+" moved to "+this.x+","+this.y);
			
		}catch(Exception e){System.out.println("AdvancedExplorerAgent - Error adding or removing position tuple.");}
		
		
	}
}